#lang typed/racket

(define-type set (-> (U set-empty set-add)))
(struct set-empty () #:transparent)
(struct set-add ([first : String] [rest : set]) #:transparent)

(: s1 set)
(define s1
  (thunk
  (set-add "a"
    (thunk
      (set-add "b"
        (thunk
          (set-add "c" (thunk (set-empty))))))))
)

(: print-set (-> set Void))
(define (print-set s)
  (match (s)
    [(set-empty)
      (void)
    ]
    [(set-add h s)
      (displayln h) ; print the first element
      (print-set s) ; print the rest
    ]
  )
)

;(print-set s1)

#|

(define (copy l)
  (match l
    [(list) (list)]
    [(list h l ...) (cons h (copy l))]
  )
)

|#


(: copy (-> set set))
(define (copy l)
  (thunk
    (match (l)
      [(set-empty) (set-empty)]
      [(set-add h s)
        (set-add h (copy s))
      ]
    )
  )
)

;(print-set (copy s1))

(: map (-> (-> String String) set set))
(define (map f l)
  (thunk
    (match (l)
      [(set-empty) (set-empty)]
      [(set-add h l)
        (define result (map f l))
        (set-add (f h) result)
      ]
    )
  )
)

;; (print-set
;;   (map string-upcase s1)
;; )

(define-type r:value r:number)
(define-type r:expression (U r:value r:apply r:variable))
(struct r:number ([value : Number]))
(struct r:variable ([name : Symbol]))
(struct r:apply (
  [func : r:expression]
  [args : (Listof r:expression)]
))
(: r:eval-builtin (-> Symbol (-> Number * Number)))
(define (r:eval-builtin sym)
  (match sym
    ['+ +]
    ['* *]
  )
)


(: r:eval-exp (-> r:expression Number))
(define (r:eval-exp exp)
  ; (r:number 10)
  (match exp
    [(r:number x) x]
    [(r:apply (r:variable f) (list arg1 arg2))
      ; f = '+
      (define racket-func (r:eval-builtin f))
      ; racket-func = +
      (define a1 (r:eval-exp arg1)) ; 3
      (define a2 (r:eval-exp arg2)) ; 10
      ;(racket-func a1 a2)
      (apply racket-func (list a1 a2))
      ; output = (+ 3 10)
    ]
  )
)

;(r:eval-exp (r:number 10))
;; (r:eval-exp
;;   (r:apply
;;     (r:variable '+)
;;     (list (r:number 3) (r:number 10))
;;   )
;; )
(: sum (-> (Listof Number) Number))
(define (sum l)
  (apply + l)
)
(sum (list 1 2 3 4))

(: f (-> Number Number Number * Void))
(define (f a b . args)
  (displayln args)
)

(f 1 2 3 4)

